#! /usr/bin/env python
# -*- coding: utf-8 -*-
####################
# Indigo Tuya plugin
# Copyright (c) 2023 Alan Carter. All rights reserved.
#

################################################################################
# Imports
################################################################################

import ipaddress
import json
import colorsys
from tuya_iot import (
    TuyaOpenAPI,
    AuthType
)


################################################################################
# Globals
################################################################################

def updateVar(name, value, folder=0):

    if name not in indigo.variables:
        indigo.variable.create(name, value=value, folder=folder)
    else:
        indigo.variable.updateValue(name, value)

################################################################################
class Plugin(indigo.PluginBase):

    ########################################
    # Class properties
    ########################################

    def __init__(self, pluginId, pluginDisplayName, pluginVersion, pluginPrefs):
     
        indigo.PluginBase.__init__(self, pluginId, pluginDisplayName, pluginVersion, pluginPrefs)
        self.userName = pluginPrefs.get("userName", "")
        self.userPass = pluginPrefs.get("userPass", "")
        self.clientId = pluginPrefs.get("clientId", "")
        self.clientSecret = pluginPrefs.get("clientSecret", "")
        self.endPoint = pluginPrefs.get("endPoint", "")
        self.logComms = pluginPrefs.get("logComms", False)
    
    ########################################
    def __del__(self):
    
        indigo.PluginBase.__del__(self)
        
    ########################################
    def startup(self):
    
        indigo.server.log("Starting Tuya plugin")
        self.openapi = TuyaOpenAPI(self.endPoint, self.clientId, self.clientSecret, AuthType.SMART_HOME)
        self.openapi.get_token()

    def deviceStartComm(self, dev):
    
        dev.stateListOrDisplayStateIdChanged()
        return


    def shutdown(self):
    
        indigo.server.log("Stopping Tuya plugin")
        pass
  
    ########################################
    
    def runConcurrentThread(self):
    
        indigo.server.log("Starting Tuya monitoring thread")
        try:
            while True:
                self.getReadings()
                self.sleep(30)
        except self.StopThread:
            self.shutdown()


    ########################################
    # Called when user prefs are changed
    
    def closedPrefsConfigUi(self, valuesDict, userCancelled):
    
        # If the user saved preferences, update logging parameters
        if userCancelled == False:
            logComms = self.logComms
            self.logComms = valuesDict.get("logComms", False)
            if self.logComms != logComms:
                if self.logComms:
                    indigo.server.log("Comms logging is on")
                else:
                    indigo.server.log("Comms logging is off")


        
    ########################################
    # Tuya specific functions
    ########################################

    def getReadings(self):
    
        for dev in indigo.devices.iter("self"):

            # Get readings      
            response = self.openapi.get("/v1.0/iot-03/devices/{}/status".format(dev.address))
            #indigo.server.log(str(response))
            if response["success"] == True:
                if self.logComms:
                    indigo.server.log("Device states for %s" % dev.name)
                    indigo.server.log(response)
                if dev.deviceTypeId == "tuyaTemp":
                    temperature = float(response["result"][0]["value"]) / 10
                    humidity = response["result"][1]["value"]
                    battery = response["result"][2]["value"]
                    unit = response["result"][3]["value"]      
                    key_value_list = [
                        {'key':'temperature', 'value':temperature, 'uiValue':str(temperature)+" "+unit.upper(), 'decimalPlaces':1},
                        {'key':'humidity', 'value':float(humidity), 'uiValue':str(humidity)+"%"},
                        {'key':'battery', 'value':battery}
                    ]       
                elif dev.deviceTypeId == "tuyaPlug":
                    onState = response["result"][0]["value"]
                    if onState:
                        dev.updateStateOnServer(key="onOffState", value=True, clearErrorState=True)
                        dev.updateStateImageOnServer(indigo.kStateImageSel.PowerOn)
                    else:
                        dev.updateStateOnServer(key="onOffState", value=False, clearErrorState=True)
                        dev.updateStateImageOnServer(indigo.kStateImageSel.PowerOff)
                    current = float(response["result"][2]["value"]) / 10
                    power = float(response["result"][3]["value"]) / 10
                    voltage = float(response["result"][4]["value"]) / 10
                    key_value_list = [
                        {'key':'current', 'value':current, 'uiValue':str(current)+" "+" A", 'decimalPlaces':1},
                        {'key':'power', 'value':float(power), 'uiValue':str(power)+" W"},
                        {'key':'voltage', 'value':float(voltage), 'uiValue':str(voltage)+" V"}
                    ]
                elif dev.deviceTypeId == "tuyaBulb":
                    onState = response["result"][0]["value"]
                    if onState:
                        dev.updateStateOnServer(key="onOffState", value=True, clearErrorState=True)
                        dev.updateStateImageOnServer(indigo.kStateImageSel.PowerOn)
                    else:
                        dev.updateStateOnServer(key="onOffState", value=False, clearErrorState=True)
                        dev.updateStateImageOnServer(indigo.kStateImageSel.PowerOff)
                    mode = response["result"][1]["value"]
                    brightness = int(response["result"][2]["value"]) / 10
                    colourTemp = int(response["result"][3]["value"])
                    key_value_list = [
                        {'key':'mode', 'value':mode},
                        {'key':"brightnessLevel", 'value':brightness, 'uiValue':str(brightness)+"%"},
                        {'key':'colour_temp', 'value':colourTemp, 'uiValue':str(colourTemp)}
                    ]
                elif dev.deviceTypeId == "tuyaPresence":
                    presence = response["result"][0]["value"]
                    sensitivity = response["result"][1]["value"]
                    near_detection = response["result"][2]["value"] 
                    far_detection = response["result"][3]["value"] 
                    target_dis_closest = response["result"][5]["value"]     
                    key_value_list = [
                        {'key':'presence_state', 'value':presence, 'uiValue':str(presence)},
                        {'key':'sensitivity', 'value':sensitivity, 'uiValue':str(sensitivity)},
                        {'key':'near_detection', 'value':near_detection, 'uiValue':str(near_detection)},
                        {'key':'far_detection', 'value':far_detection, 'uiValue':str(far_detection)},
                        {'key':'target_dis_closest', 'value':target_dis_closest, 'uiValue':str(target_dis_closest)}
                    ]
                    if presence == "presence":
                        dev.updateStateImageOnServer(indigo.kStateImageSel.MotionSensorTripped)
                    else:
                        dev.updateStateImageOnServer(indigo.kStateImageSel.MotionSensor)
                elif dev.deviceTypeId == "tuyaSmartIr":
                    data = {"categoryId": "infrared_tv", "key": "Power"}
                    #response = self.openapi.post("/v2.0/infrareds/{}/remotes/bf4013f3f74b6412f66ryy/command?categoryId=infrared_tv&key=Power".format(dev.address), data)
                    #indigo.server.log(str(response))
                dev.updateStatesOnServer(key_value_list)
            else:
                indigo.server.log("Error getting readings for Tuya device", isError=True)
                indigo.server.log(str(response))

            # Get info
            response = self.openapi.get("/v1.1/iot-03/devices/{}".format(dev.address))
            if response["success"] == True:
                if self.logComms:
                    indigo.server.log("Device info for %s" % dev.name)
                    indigo.server.log(response)     
                local_key = response["result"]["local_key"]
                category = response["result"]["category_name"]
                model = response["result"]["model"]
                name = response["result"]["name"]
                product = response["result"]["product_name"]
                if response["result"]["online"] == True:
                    status = "Online"
                else:
                    status = "Offline"
        
                key_value_list = [
                    {'key':'local_key', 'value':local_key},
                    {'key':'category', 'value':category},
                    {'key':'model', 'value':model},
                    {'key':'name', 'value':name},
                    {'key':'product', 'value':product},
                    {'key':'status', 'value':status},
                ]
                dev.updateStatesOnServer(key_value_list)
  
            else:
                indigo.server.log("Error getting info for Tuya device", isError=True)
                indigo.server.log(str(response))



    #######################################
    # Action functions
    #######################################
    
    def setWhiteMode(self, pluginAction, actionId):
        devToProcess = int(pluginAction.props["bulb"])
        device = None
        for dev in indigo.devices.iter("self"):
            if dev.id == devToProcess:
                device = indigo.devices[dev.name]
                if device != None:
                    if dev.deviceTypeId == "tuyaBulb":
                        data = {"commands":[{"code":"work_mode","value":"white"}]}
                        try:
                            response = self.openapi.post("/v1.0/iot-03/devices/{}/commands".format(dev.address), data)
                            if response["success"] == True:
                                dev.updateStateOnServer("mode", "white")
                                return True
                            else:
                                indigo.server.log("%s mode -> white command failed" % dev.name, isError=True)
                        except:
                            indigo.server.log("%s mode -> white command failed" % dev.name, isError=True)          
                    else:
                        indigo.server.log("Colour commands can only be sent to colour bulb devices", isError=True)
            

    def setColourMode(self, pluginAction, actionId):
        devToProcess = int(pluginAction.props["bulb"])
        device = None
        for dev in indigo.devices.iter("self"):
            if dev.id == devToProcess:
                device = indigo.devices[dev.name]
                if device != None:
                    if dev.deviceTypeId == "tuyaBulb":
                        data = {"commands":[{"code":"work_mode","value":"colour"}]}
                        try:
                            response = self.openapi.post("/v1.0/iot-03/devices/{}/commands".format(dev.address), data)
                            if response["success"] == True:
                                dev.updateStateOnServer("mode", "colour")
                                return True
                            else:
                                indigo.server.log("%s mode -> colour command failed" % dev.name, isError=True)
                        except:
                            indigo.server.log("%s mode -> colour command failed" % dev.name, isError=True)          
                    else:
                        indigo.server.log("Colour commands can only be sent to colour bulb devices", isError=True)  


    def sendIrCommand(self, pluginAction):
    
        ir_transmitter_device = indigo.devices[int(pluginAction.props["ir_transmitter"])]
        ir_transmitter = ir_transmitter_device.address
        ir_remote_device = indigo.devices[int(pluginAction.props["ir_remote"])]
        ir_remote = ir_remote_device.address
        ir_command = pluginAction.props["ir_command"]
        data = {"categoryId": "infrared_tv", "key": ir_command}
        if ir_command == "mute":
        	raw = "raw/"
        else:
        	raw = ""
        try:
            response = self.openapi.post("/v2.0/infrareds/{}/remotes/{}/{}command".format(ir_transmitter, ir_remote, raw), data)
            if not response['success']:
                indigo.server.log("Command error", isError=True)
                indigo.server.log(str(response))
            else:
                return True
        except:
            indigo.server.log("Command fail", isError=True)
            indigo.server.log(str(response))

 

    ########################################
    # Smart device action callback
    ########################################
    
    # Main switch action bottleneck called by Indigo Server.
    def actionControlDevice(self, action, dev):

        if action.deviceAction == indigo.kDeviceAction.TurnOn:
            if  dev.deviceTypeId == "tuyaPlug":
                data = {"commands":[{"code":"switch","value":True}]}
            elif dev.deviceTypeId == "tuyaBulb":
                data = {"commands":[{"code":"switch_led","value":True}]}
            try:
                response = self.openapi.post("/v1.0/iot-03/devices/{}/commands".format(dev.address), data)
                if response["success"] == True:
                    dev.updateStateOnServer("onOffState", True)
                    dev.updateStateImageOnServer(indigo.kStateImageSel.PowerOn)
                else:
                    indigo.server.log("%s 'on' command failed" % dev.name, isError=True)
            except:
                 indigo.server.log("%s 'on' command failed" % dev.name, isError=True)
                
        elif action.deviceAction == indigo.kDeviceAction.TurnOff:
            if  dev.deviceTypeId == "tuyaPlug":
                data = {"commands":[{"code":"switch","value":False}]}
            elif dev.deviceTypeId == "tuyaBulb":
                data = {"commands":[{"code":"switch_led","value":False}]}
            try:
                response = self.openapi.post("/v1.0/iot-03/devices/{}/commands".format(dev.address), data)
                if response["success"] == True:
                    dev.updateStateOnServer("onOffState", False)
                    dev.updateStateImageOnServer(indigo.kStateImageSel.PowerOff)
                else:
                    indigo.server.log("%s 'off' command failed" % dev.name, isError=True)
            except:
                indigo.server.log("%s 'off' command failed" % dev.name, isError=True)

        elif action.deviceAction == indigo.kDeviceAction.SetBrightness:
            if dev.deviceTypeId == "tuyaBulb":
                new_brightness = action.actionValue
                if new_brightness < 1:
                    new_brightness = 1
                scaled_brightness = new_brightness * 10
                data = {"commands":[{"code":"bright_value_v2","value":scaled_brightness}]}
                try:
                    response = self.openapi.post("/v1.0/iot-03/devices/{}/commands".format(dev.address), data)
                    if response["success"] == True:
                        dev.updateStateOnServer(key="brightnessLevel", value=new_brightness, uiValue=str(new_brightness)+"%")
                    else:
                        indigo.server.log(f"Command \"{dev.name}\" set brightness to {new_brightness} failed", isError=True)
                except:
                    indigo.server.log(f"Command \"{dev.name}\" set brightness to {new_brightness} error", isError=True)
            else:
                indigo.server.log("Brightness commands can only be sent to bulb devices", isError=True)

        elif action.deviceAction == indigo.kDeviceAction.BrightenBy:
            if dev.deviceTypeId == "tuyaBulb":
                new_brightness = min(dev.brightness + action.actionValue, 100)
                if new_brightness < 1:
                	new_brightness = 1
                scaled_brightness = new_brightness * 10
                data = {"commands":[{"code":"bright_value_v2","value":scaled_brightness}]}
                try:
                    response = self.openapi.post("/v1.0/iot-03/devices/{}/commands".format(dev.address), data)
                    if response["success"] == True:
                        dev.updateStateOnServer(key="brightnessLevel", value=new_brightness, uiValue=str(new_brightness)+"%")
                    else:
                        indigo.server.log(f"Command \"{dev.name}\" brighten to {new_brightness} failed", isError=True)
                except:
                    indigo.server.log(f"Command \"{dev.name}\" brighten to {new_brightness} error", isError=True)
            else:
                indigo.server.log("Brightness commands can only be sent to bulb devices", isError=True)

        elif action.deviceAction == indigo.kDeviceAction.DimBy:
            if dev.deviceTypeId == "tuyaBulb":
                new_brightness = max(dev.brightness - action.actionValue, 0)
                if new_brightness < 1:
                	new_brightness = 1
                scaled_brightness = new_brightness * 10
                data = {"commands":[{"code":"bright_value_v2","value":scaled_brightness}]}
                try:
                    response = self.openapi.post("/v1.0/iot-03/devices/{}/commands".format(dev.address), data)
                    if response["success"] == True:
                        dev.updateStateOnServer(key="brightnessLevel", value=new_brightness, uiValue=str(new_brightness)+"%")
                    else:
                        indigo.server.log(f"Command \"{dev.name}\" dim to {new_brightness} failed", isError=True)
                except:
                    indigo.server.log(f"Command \"{dev.name}\" dim to {new_brightness} error", isError=True)
            else:
                indigo.server.log("Brightness commands can only be sent to bulb devices", isError=True)
                

        elif action.deviceAction == indigo.kDimmerRelayAction.SetColorLevels:
            if dev.deviceTypeId == "tuyaBulb":
                isError = False
                if dev.pluginProps.get('SupportsRGB', False):
                    if "blueLevel" in str(action.actionValue):
                        red = action.actionValue.get('redLevel', 0)
                        if red == "":
                            red = 0
                            valuesDict['redLevel'] = red
                        green = action.actionValue.get('greenLevel', 0)
                        if green == "":
                            green = 0
                            valuesDict['greenLevel'] = green
                        blue = action.actionValue.get('blueLevel', 0)
                        if blue == "":
                            blue = 0
                            valuesDict['blueLevel'] = blue

                        # Validate red value.
                        try:
                            red = int(red)
                            if (red < 0) or (red > 100):
                                isError = True
                                errorsDict['red'] = "Red values must be a whole number between 0 and 100."
                                errorsDict['showAlertText'] += errorsDict['red'] + "\n\n"
                        except ValueError:
                            isError = True
                            errorsDict['red'] = "Red values must be a whole number between 0 and 100."
                            errorsDict['showAlertText'] += errorsDict['red'] + "\n\n"
                        except Exception as e:
                            isError = True
                            errorsDict['red'] = "Invalid Red value: {}  @line#:{})".format(e,sys.exc_info()[2].tb_lineno)
                            errorsDict['showAlertText'] += errorsDict['red'] + "\n\n"

                        # Validate green value.
                        try:
                            green = int(green)
                            if (green < 0) or (green > 100):
                                isError = True
                                errorsDict['green'] = "Green values must be a whole number between 0 and 100."
                                errorsDict['showAlertText'] += errorsDict['green'] + "\n\n"
                        except ValueError:
                            isError = True
                            errorsDict['green'] = "Green values must be a whole number between 0 and 100."
                            errorsDict['showAlertText'] += errorsDict['green'] + "\n\n"
                        except Exception as e:
                            isError = True
                            errorsDict['green'] = "Invalid Green value: {}  @line#:{})".format(e,sys.exc_info()[2].tb_lineno)
                            errorsDict['showAlertText'] += errorsDict['green'] + "\n\n"

                        # Validate blue value.
                        try:
                            blue = int(blue)
                            if (blue < 0) or (blue > 100):
                                isError = True
                                errorsDict['blue'] = "Blue values must be a whole number between 0 and 100."
                                errorsDict['showAlertText'] += errorsDict['blue'] + "\n\n"
                        except ValueError:
                            isError = True
                            errorsDict['blue'] = "Blue values must be a whole number between 0 and 100."
                            errorsDict['showAlertText'] += errorsDict['blue'] + "\n\n"
                        except Exception as e:
                            isError = True
                            errorsDict['blue'] = "Invalid Blue value: {}  @line#:{})".format(e,sys.exc_info()[2].tb_lineno)
                            errorsDict['showAlertText'] += errorsDict['blue'] + "\n\n"

                        if not isError:
                            hVal, sVal, vVal = colorsys.rgb_to_hsv(red, green, blue)
                            #indigo.server.log("h: %d, s: %d, v: %d" % (hVal*360, sVal*1000, vVal*10))
                            data = {"commands":[{"code":"colour_data_v2", "value":{"h":int(hVal*360), "s":int(sVal*1000), "v":int(vVal*10)}}]}
                            try:
                                response = self.openapi.post("/v1.0/iot-03/devices/{}/commands".format(dev.address), data)
                                if response["success"] == True:
                                    indigo.server.log("Success!")
                                    #dev.updateStateOnServer(key="brightnessLevel", value=new_brightness, uiValue=str(new_brightness)+"%")
                                else:
                                    indigo.server.log(f"Command \"{dev.name}\" failed", isError=True)
                            except:
                                indigo.server.log(f"Command \"{dev.name}\" error", isError=True)
                                
                    elif "whiteLevel" in str(action.actionValue):
                        if dev.deviceTypeId == "tuyaBulb":
                            new_brightness = action.actionValue.get('whiteLevel', 0)
                            if new_brightness < 1:
                                new_brightness = 1
                            scaled_brightness = int(new_brightness * 10)
                            data = {"commands":[{"code":"bright_value_v2","value":scaled_brightness}]}
                            try:
                                response = self.openapi.post("/v1.0/iot-03/devices/{}/commands".format(dev.address), data)
                                if response["success"] == True:
                                    dev.updateStateOnServer(key="whiteLevel", value=new_brightness, uiValue=str(new_brightness)+"%")
                                else:
                                    indigo.server.log(str(response))
                                    indigo.server.log(f"Command \"{dev.name}\" set white level to {new_brightness} failed", isError=True)
                            except:
                                indigo.server.log(f"Command \"{dev.name}\" set white level to {new_brightness} error", isError=True)

                    
                    elif "whiteTemperature" in str(action.actionValue):
                        if dev.deviceTypeId == "tuyaBulb":
                            isError = False
                            whiteTemp = action.actionValue.get('whiteTemperature', 0)
                            try:
                                whiteTemp = int(whiteTemp)
                                if (whiteTemp < 2200) or (whiteTemp > 6500):
                                    isError = True
                                    errorsDict['whiteTemperature'] = "White Temperature values must be a whole number between 2200 and 6500."
                                    errorsDict['showAlertText'] += errorsDict['whiteTemperature'] + "\n\n"
                            except ValueError:
                                isError = True
                                errorsDict['whiteTemperature'] = "White Temperature values must be a whole number between 2200 and 6500."
                                errorsDict['showAlertText'] += errorsDict['whiteTemperature'] + "\n\n"
                            except Exception as e:
                                isError = True
                                errorsDict['whiteTemperature'] = "Invalid White Temperature value: {}  @line#:{})".format(e,sys.exc_info()[2].tb_lineno)
                                errorsDict['showAlertText'] += errorsDict['whiteTemperature'] + "\n\n"
                            whiteTempScaled = int((whiteTemp - 2200) / 4.3)
                            if not isError:
                                data = {"commands":[{"code":"temp_value_v2","value":whiteTempScaled}]}
                                try:
                                    response = self.openapi.post("/v1.0/iot-03/devices/{}/commands".format(dev.address), data)
                                    if response["success"] == True:
                                        dev.updateStateOnServer(key="whiteTemperature", value=whiteTemp, uiValue=str(whiteTemp))
                                    else:
                                        indigo.server.log(str(response))
                                        indigo.server.log(f"Command \"{dev.name}\" set temperature to {whiteTemp} failed", isError=True)
                                except:
                                    indigo.server.log(f"Command \"{dev.name}\" set temperature to {whiteTemp} error", isError=True)
                        
            else:
                indigo.server.log("Colour commands can only be sent to colour bulb devices", isError=True)
                

